@@ -40,13 +40,17 @@
 #include <stddef.h>
 #include <errno.h>
 
+#ifdef HAVE_ICONV
+#include <iconv.h>
+#endif
+
 #include <memory.h>
-#include <math.h>
 #include <sys/types.h>
 #include <string.h>
 #include <wchar.h>
 
 #include "../include/libxls/endian.h"
+#include "../include/libxls/locale.h"
 #include "../include/xls.h"
 
 #ifndef min
@@ -62,7 +66,7 @@ static xls_formula_handler formula_handler;
 static xls_error_t xls_addSST(xlsWorkBook* pWB, SST* sst, DWORD size);
 static xls_error_t xls_appendSST(xlsWorkBook* pWB, BYTE* buf, DWORD size);
 static xls_error_t xls_addFormat(xlsWorkBook* pWB, FORMAT* format, DWORD size);
-static char* xls_addSheet(xlsWorkBook* pWB, BOUNDSHEET* bs, DWORD size);
+static xls_error_t xls_addSheet(xlsWorkBook* pWB, BOUNDSHEET* bs, DWORD size);
 static xls_error_t xls_addRow(xlsWorkSheet* pWS,ROW* row);
 static xls_error_t xls_makeTable(xlsWorkSheet* pWS);
 static struct st_cell_data *xls_addCell(xlsWorkSheet* pWS, BOF* bof, BYTE* buf);
@@ -222,27 +226,26 @@ static xls_error_t xls_appendSST(xlsWorkBook* pWB, BYTE* buf, DWORD size)
             if (flag & 0x1) {
                 size_t new_len = 0;
                 ln_toread = min((size-ofs)/2, ln);
-                ret=unicode_decode((char *)buf+ofs,ln_toread*2,&new_len,pWB->charset);
+                ret=unicode_decode((char *)buf+ofs, ln_toread*2, pWB);
 
-                if (ret == NULL)
-                {
+                if (ret == NULL) {
                     ret = strdup("*failed to decode utf16*");
-                    new_len = strlen(ret);
                 }
 
-                ret = realloc(ret,new_len+1);
-                ret[new_len]=0;
-
                 ln -= ln_toread;
                 ofs+=ln_toread*2;
 
                 if (xls_debug) {
+                    new_len = strlen(ret);
 	                printf("String16SST: %s(%lu)\n", ret, (unsigned long)new_len);
                 }
             } else {
                 ln_toread = min((size-ofs), ln);
 
-				ret = utf8_decode((char *)buf+ofs, ln_toread, pWB->charset);
+                ret = codepage_decode((char *)buf+ofs, ln_toread, pWB);
+                if (ret == NULL) {
+                    ret = strdup("*failed to decode BIFF5 string*");
+                }
 
                 ln  -= ln_toread;
                 ofs += ln_toread;
@@ -341,7 +344,7 @@ static double NumFromRk(DWORD drk)
     return ret;
 }
 
-static char * xls_addSheet(xlsWorkBook* pWB, BOUNDSHEET *bs, DWORD size)
+static xls_error_t xls_addSheet(xlsWorkBook* pWB, BOUNDSHEET *bs, DWORD size)
 {
 	char * name;
 	DWORD filepos;
@@ -353,7 +356,7 @@ static char * xls_addSheet(xlsWorkBook* pWB, BOUNDSHEET *bs, DWORD size)
 
 	// printf("charset=%s uni=%d\n", pWB->charset, unicode);
 	// printf("bs name %.*s\n", bs->name[0], bs->name+1);
-	name = get_string(bs->name, size - offsetof(BOUNDSHEET, name), 0, pWB->is5ver, pWB->charset);
+	name = get_string(bs->name, size - offsetof(BOUNDSHEET, name), 0, pWB);
 	// printf("name=%s\n", name);
 
 	if(xls_debug) {
@@ -388,7 +391,7 @@ static char * xls_addSheet(xlsWorkBook* pWB, BOUNDSHEET *bs, DWORD size)
 
     pWB->sheets.sheet = realloc(pWB->sheets.sheet,(pWB->sheets.count+1)*sizeof (struct st_sheet_data));
     if (pWB->sheets.sheet == NULL)
-        return NULL;
+        return LIBXLS_ERROR_MALLOC;
 
     pWB->sheets.sheet[pWB->sheets.count].name=name;
     pWB->sheets.sheet[pWB->sheets.count].filepos=filepos;
@@ -396,7 +399,7 @@ static char * xls_addSheet(xlsWorkBook* pWB, BOUNDSHEET *bs, DWORD size)
     pWB->sheets.sheet[pWB->sheets.count].type=type;
     pWB->sheets.count++;
 
-	return name;
+	return LIBXLS_OK;
 }
 
 
@@ -469,7 +472,7 @@ int xls_isCellTooSmall(xlsWorkBook* pWB, BOF* bof, BYTE* buf) {
     if (bof->id == XLS_RECORD_LABELSST)
         return (bof->size < offsetof(LABEL, value) + (pWB->is5ver ? 2 : 4));
 
-    if (bof->id == XLS_RECORD_LABEL) {
+    if (bof->id == XLS_RECORD_LABEL || bof->id == XLS_RECORD_RSTRING) {
         if (bof->size < offsetof(LABEL, value) + 2)
             return 1;
 
@@ -596,6 +599,7 @@ static struct st_cell_data *xls_addCell(xlsWorkSheet* pWS,BOF* bof,BYTE* buf)
         break;
     case XLS_RECORD_LABELSST:
     case XLS_RECORD_LABEL:
+    case XLS_RECORD_RSTRING:
         xls_cell_set_str(cell, xls_getfcell(pWS->workbook, cell, ((LABEL*)buf)->value));
         if (cell->str) {
             sscanf((char *)cell->str, "%d", &cell->l);
@@ -642,7 +646,7 @@ static char *xls_addFont(xlsWorkBook* pWB, FONT* font, DWORD size)
 
     tmp=&pWB->fonts.font[pWB->fonts.count];
 
-    tmp->name = get_string(font->name, size - offsetof(FONT, name), 0, pWB->is5ver, pWB->charset);
+    tmp->name = get_string(font->name, size - offsetof(FONT, name), 0, pWB);
 
     tmp->height=font->height;
     tmp->flag=font->flag;
@@ -670,7 +674,7 @@ static xls_error_t xls_addFormat(xlsWorkBook* pWB, FORMAT* format, DWORD size)
 
     tmp = &pWB->formats.format[pWB->formats.count];
     tmp->index = format->index;
-    tmp->value = get_string(format->value, size - offsetof(FORMAT, value), (BYTE)!pWB->is5ver, (BYTE)pWB->is5ver, pWB->charset);
+    tmp->value = get_string(format->value, size - offsetof(FORMAT, value), (BYTE)!pWB->is5ver, pWB);
     if(xls_debug) xls_showFormat(tmp);
     pWB->formats.count++;
 
@@ -873,7 +877,6 @@ xls_error_t xls_parseWorkBook(xlsWorkBook* pWB)
         case XLS_RECORD_BOF:	// BIFF5-8
             pWB->is5ver = (buf[0] + (buf[1] << 8) != 0x600);
             pWB->type = buf[2] + (buf[3] << 8);
-
             if(xls_debug) {
                 printf("version: %s\n", pWB->is5ver ? "BIFF5" : "BIFF8" );
                 printf("   type: %.2X\n", pWB->type);
@@ -882,7 +885,7 @@ xls_error_t xls_parseWorkBook(xlsWorkBook* pWB)
 
         case XLS_RECORD_CODEPAGE:
             pWB->codepage = buf[0] + (buf[1] << 8);
-			if(xls_debug) printf("codepage=%x\n", pWB->codepage);
+			if(xls_debug) printf("codepage: %d\n", pWB->codepage);
             break;
 
         case XLS_RECORD_CONTINUE:
@@ -931,9 +934,10 @@ xls_error_t xls_parseWorkBook(xlsWorkBook* pWB)
 				//printf("ADD SHEET\n");
 				BOUNDSHEET *bs = (BOUNDSHEET *)buf;
                 xlsConvertBoundsheet(bs);
-				//char *s;
 				// different for BIFF5 and BIFF8
-				/*s = */ xls_addSheet(pWB, bs, bof1.size);
+                if ((retval = xls_addSheet(pWB, bs, bof1.size)) != LIBXLS_OK) {
+                    goto cleanup;
+                }
 			}
             break;
 
@@ -1009,7 +1013,7 @@ xls_error_t xls_parseWorkBook(xlsWorkBook* pWB)
 					printf("  ident: 0x%x\n", styl->ident);
 					printf("  level: 0x%x\n", styl->lvl);
 				} else {
-					char *s = get_string((char *)&buf[2], bof1.size - 2, 1, pWB->is5ver, pWB->charset);
+					char *s = get_string((char *)&buf[2], bof1.size - 2, 1, pWB);
 					printf("  name=%s\n", s);
                     free(s);
 				}
@@ -1124,8 +1128,9 @@ static xls_error_t xls_preparseWorkSheet(xlsWorkSheet* pWS)
                 goto cleanup;
             }
             xlsConvertRow((ROW*)buf);
-            if (pWS->rows.lastcol<((ROW*)buf)->lcell)
-                pWS->rows.lastcol=((ROW*)buf)->lcell;
+            /* The lcell field is 1-indexed whereas lastcol is 0-indexed */
+            if (pWS->rows.lastcol+1<((ROW*)buf)->lcell)
+                pWS->rows.lastcol=((ROW*)buf)->lcell-1;
             if (pWS->rows.lastrow<((ROW*)buf)->index)
                 pWS->rows.lastrow=((ROW*)buf)->index;
             break;
@@ -1156,6 +1161,7 @@ static xls_error_t xls_preparseWorkSheet(xlsWorkSheet* pWS)
         case XLS_RECORD_LABELSST:
         case XLS_RECORD_BLANK:
         case XLS_RECORD_LABEL:
+        case XLS_RECORD_RSTRING:
         case XLS_RECORD_FORMULA:
         case XLS_RECORD_FORMULA_ALT:
         case XLS_RECORD_BOOLERR:
@@ -1169,7 +1175,7 @@ static xls_error_t xls_preparseWorkSheet(xlsWorkSheet* pWS)
                 pWS->rows.lastrow=xlsShortVal(((COL*)buf)->row);
             break;
         }
-        if (pWS->rows.lastcol > 256) {
+        if (pWS->rows.lastcol > 255) {
             retval = LIBXLS_ERROR_PARSE;
             goto cleanup;
         }
@@ -1342,6 +1348,7 @@ xls_error_t xls_parseWorkSheet(xlsWorkSheet* pWS)
         case XLS_RECORD_LABELSST:
         case XLS_RECORD_BLANK:
         case XLS_RECORD_LABEL:
+        case XLS_RECORD_RSTRING:
         case XLS_RECORD_FORMULA:
         case XLS_RECORD_FORMULA_ALT:
             if ((cell = xls_addCell(pWS, &tmp, buf)) == NULL) {
@@ -1356,7 +1363,7 @@ xls_error_t xls_parseWorkSheet(xlsWorkSheet* pWS)
 		case XLS_RECORD_STRING:
 			if(cell && (cell->id == XLS_RECORD_FORMULA || cell->id == XLS_RECORD_FORMULA_ALT)) {
                 xls_cell_set_str(cell, get_string((char *)buf, tmp.size,
-                            (BYTE)!pWB->is5ver, pWB->is5ver, pWB->charset));
+                            (BYTE)!pWB->is5ver, pWB));
 				if (xls_debug) xls_showCell(cell);
 			}
 			break;
@@ -1456,12 +1463,7 @@ static xlsWorkBook *xls_open_ole(OLE2 *ole, const char *charset, xls_error_t *ou
     pWB->sheets.count=0;
     pWB->xfs.count=0;
     pWB->fonts.count=0;
-    if (charset) {
-        pWB->charset = malloc(strlen(charset) * sizeof(char)+1);
-        strcpy(pWB->charset, charset);
-    } else {
-        pWB->charset = strdup("UTF-8");
-    }
+    pWB->charset = strdup(charset ? charset : "UTF-8");
 
     retval = xls_parseWorkBook(pWB);
 
@@ -1596,8 +1598,18 @@ void xls_close_WB(xlsWorkBook* pWB)
 	if(pWB->summary)  free(pWB->summary);
 	if(pWB->docSummary) free(pWB->docSummary);
 
-	// TODO - free other dynamically allocated objects like string table??
-	free(pWB);
+#ifdef HAVE_ICONV
+    if (pWB->converter)
+        iconv_close((iconv_t)pWB->converter);
+    if (pWB->utf16_converter)
+        iconv_close((iconv_t)pWB->utf16_converter);
+#endif
+
+    if (pWB->utf8_locale)
+        xls_freelocale((xls_locale_t)pWB->utf8_locale);
+
+    // TODO - free other dynamically allocated objects like string table??
+    free(pWB);
 }
 
 void xls_close_WS(xlsWorkSheet* pWS)
